// SpamWatchClient.cpp: implementation of the SpamWatchClient class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SpamWatchClient.h"
#include <boost/regex.hpp>
#include <time.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


	// Connect to the server, return 0 if successful, 
	// return non-zero value if connection cannot be established
	// result_handler is the handler called when result for checking a text becomes available
int SpamWatchClient::Connect(LPTSTR host, const int port, void (*result_handler)(int seq, int result, void *arg))
{
	USES_CONVERSION;

	m_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	struct sockaddr_in saServer;
	saServer.sin_family = AF_INET;
	saServer.sin_addr.s_addr = inet_addr(T2A(host));
	saServer.sin_port = htons(port);

	int nRet = connect(m_socket, (struct sockaddr *)&saServer, sizeof(struct sockaddr));
	
	m_result_handler = result_handler;

	return nRet;
}

// Check whether text is the content of a spam email.
// The text will be normalized before checking.
// The returned value is a sequence number of the query
// The sequence number will also be passed to the result_handler when result
// for this query is available
int SpamWatchClient::CheckText(LPTSTR text, void *arg)
{
	USES_CONVERSION;
	LPTSTR szNormalized = Normalize(text);
	unsigned fps[10];
	LPSTR byteContent = T2A(szNormalized);
	Digest(byteContent, fps);
	free(szNormalized);

	char buf[1024];
	sprintf(buf, "%d check %u %u %u %u %u %u %u %u %u %u\n", m_next_seq, 
		fps[0], fps[1], fps[2], fps[3], fps[4], fps[5], 
		fps[6], fps[7], fps[8], fps[9]);
	send(m_socket, buf, strlen(buf), 0);

	spam_request r;
	r.sequence = m_next_seq;
	r.type = REQUEST_CHECK;
	r.arg = arg;

	m_requests.push_back(r);

	ProcessReply();
	return m_next_seq++;
}

// Vote for or against the text.  Tell the server whether the specified
// text is or is not spam.
// vote can be NOT_SPAM or SPAM
int SpamWatchClient::Vote(LPTSTR text, const int vote)
{
	USES_CONVERSION;
	LPTSTR szNormalized = Normalize(text);
	unsigned fps[10];
	LPSTR byteContent = T2A(szNormalized);
	Digest(byteContent, fps);
	free(szNormalized);
	
	char buf[1024];
	sprintf(buf, "%d vote %d %u %u %u %u %u %u %u %u %u %u\n", m_next_seq, vote,
		fps[0], fps[1], fps[2], fps[3], fps[4], fps[5], 
		fps[6], fps[7], fps[8], fps[9]);
	send(m_socket, buf, strlen(buf), 0);

	spam_request r;
	r.sequence = m_next_seq;
	r.type = REQUEST_VOTE;

	m_requests.push_back(r);

	ProcessReply();
	return m_next_seq++;
}
	
// Wait for all the queries to finish and Closes the connection
void SpamWatchClient::Close()
{
	// wait for all queries
	Join(10000);

	char buf[50];
	sprintf(buf, "%d quit\n", m_next_seq++);
	int nRet = send(m_socket, buf, strlen(buf), 0);
	shutdown(m_socket, SD_SEND);
	closesocket(m_socket);
}

// Remove all redundant space characters from the text
LPTSTR SpamWatchClient::Normalize(LPTSTR text)
{
	USES_CONVERSION;
	boost::RegEx expr("\\s+");

	// convert the input string to ASCII
	LPSTR spText = T2A(text);
	std::string sText(spText);
	std::string sResult = expr.Merge(sText, " ");

	// convert the result back to UNICODE
	LPTSTR spResult = A2T((LPSTR)sResult.c_str());
	
	return _tcsdup(spResult);
}


// Calculate fingerprints for the text.
// [IN] content - byte array of content to calculate finger prints for
// [OUT] fps - pointer to buffer to put fingerprints
// Return 0 if success.  Return 1 if content is too short.  2 if not enough
// fingerprints can be found.
int SpamWatchClient::Digest(LPSTR content, unsigned int *fps)
{
	int len = strlen(content);

	if (len < FINGERPRINTING_LENGTH + FINGERPRINT_COUNT - 1)
		return 1;
	int digestLength = len - FINGERPRINTING_LENGTH + 1;
	unsigned * digests = (unsigned *) malloc(digestLength * sizeof(unsigned));

	unsigned d;
	
	// Calc first digest
	d = 0;
	for (int i = 0; i < FINGERPRINTING_LENGTH; i++) {
		d = (d * PRIME) + (unsigned)(content[i] & 0xff);
	}
	digests[0] = d;

	// calc other digests based on the previous one
	for (i = FINGERPRINTING_LENGTH; i < len; i++) {
		digests[i - FINGERPRINTING_LENGTH + 1] =
			(digests[i - FINGERPRINTING_LENGTH] * PRIME
				+ (unsigned)(content[i] & 0xff)
				+ (unsigned)(content[i - FINGERPRINTING_LENGTH] & 0xff)
				* ptoNComplement);
/*
		d = 0;
		for (int j = 0; j < FINGERPRINTING_LENGTH; j++) {
			d = (d * PRIME) + (unsigned)(content[i+j-FINGERPRINTING_LENGTH+1] & 0xff);
		}
	
		if (d != digests[i - FINGERPRINTING_LENGTH + 1]) {
			MessageBox(NULL, "Not Equal!", "Error", MB_OK);
		}
*/
	}


	// Select some block digests as fingerprints
	bool *taken = (bool *)malloc(digestLength * sizeof(bool));
	for (i = 0; i < digestLength; i++)
		taken[i] = false;

	unsigned lastChosen = (unsigned)0xffffffff;		// used to avoid duplication
	for (i = 0; i < FINGERPRINT_COUNT; i++) {
		unsigned chosen = 0;
		int chosenIndex = -1;
		for (int j = 0; j < digestLength; j++) {
			if (digests[j] >= chosen && !taken[j] && digests[j] < lastChosen) {
				chosen = digests[j];
				chosenIndex = j;
			}
		}
		if (chosenIndex == -1)
			return 2;
		fps[i] = chosen;
		lastChosen = chosen;
		taken[chosenIndex] = true;
	}	

	free(digests);
	free(taken);
	return 0;
}


void SpamWatchClient::Join(int timeout)
{
	time_t start, now;
	time(&start);

	while (!m_requests.empty()) {
		bool bData = ProcessReply();
		time(&now);
		if (difftime(now, start) > timeout / 1000.0)
			return;

		if (!bData)
			Sleep(100);		// Wait some time before try again
	}
}

bool SpamWatchClient::ProcessReply()
{
	unsigned long n;
	char line[1025];
	int line_length;
	bool result = false;

	while (true) {
		// peek to see if data is available
		ioctlsocket(m_socket, FIONREAD, &n);

		if (n == 0)
			return result;

		result = true;

		// Read data into the buffer and process them in lines
		// The buffer may contain leftover data from last time
		int len = recv(m_socket, m_recv_buf + m_recv_buf_pos, 1024 - m_recv_buf_pos, 0);
		len += m_recv_buf_pos;

		int start = 0, cur;
		while (true) {
			for (cur = start; cur < len && m_recv_buf[cur] != '\n'; cur++) {}
			if (cur == len) {
				// The last line is not complete
				// Move the incomplete line to the beginning
				memmove(m_recv_buf, m_recv_buf + start, cur - start);
				m_recv_buf_pos = cur - start;
	
				if (m_recv_buf_pos == 1024)		// That's a line more than 1024
					m_recv_buf_pos = 0;			// Then just drop it !!!
				
				break;
			}

			line_length = cur - start;
			memmove(line, m_recv_buf + start, line_length);	// move the line excluding \n
			line[line_length] = '\0';

			ProcessReplyLine(line, line_length);

			// deal with \n\r
			if ((cur + 1) < len ) {
				if (m_recv_buf[cur+1] == '\r')
					start = cur + 2;
				else
					start = cur + 1;
			} else {
				// Finished
				m_recv_buf_pos = 0;
				break;
			}
		}
	}
	return result;
}

// Parse the reply line and process it
void SpamWatchClient::ProcessReplyLine(LPSTR line, int len) 
{
	if (m_requests.empty())
		return;

	// The next request
	spam_request &r_ref = m_requests.front();
	spam_request r = r_ref;
	m_requests.erase(m_requests.begin());
	
	int spc = strcspn(line, " ");
	line[spc] = '\0';

	int seq = atoi(line);
	
	if (seq != r.sequence) {
//		printf("Unmatched sequence!");
		return;
	}

	if (r.type == REQUEST_CHECK) {
		char *line2 = line + spc + 1;
		int spc_status = strcspn(line2, " ");
		line2[spc_status] = '\0';

		int status = atoi(line2);

		if (status != 200) {
//			printf("Status not 200!");
			return;
		}

		char *line3 = line2 + spc_status + 1;
		int result = atoi(line3);

		m_result_handler(seq, result, r.arg);
	}
}